import 'dart:async';
import 'dart:convert';

import 'package:flutter/foundation.dart';
import 'package:flutter/gestures.dart';
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';
import 'package:flutter/services.dart';
import 'package:stream_transform/stream_transform.dart';

import '../helpers/events.dart';
import '../helpers/misc.dart';
import '../helpers/types.dart';
import 'cocos_widget_platform.dart';
import 'windows_cocos_widget_view.dart';

class MethodChannelCocosWidget extends CocosWidgetPlatform {
  // Every method call passes the int cocosId
  late final Map<int, MethodChannel> _channels = {};

  /// Set [CocosWidgetFlutterPlatform] to use [AndroidViewSurface] to build the Google Maps widget.
  ///
  /// This implementation uses hybrid composition to render the Cocos Widget
  /// Widget on Android. This comes at the cost of some performance on Android
  /// versions below 10. See
  /// https://flutter.dev/docs/development/platform-integration/platform-views#performance for more
  /// information.
  /// Defaults to false.
  bool useAndroidViewSurface = true;

  /// Accesses the MethodChannel associated to the passed cocosId.
  MethodChannel channel(int cocosId) {
    MethodChannel? channel = _channels[cocosId];
    if (channel == null) {
      throw UnknownCocosIDError(cocosId);
    }
    return channel;
  }

  MethodChannel ensureChannelInitialized(int cocosId) {
    MethodChannel? channel = _channels[cocosId];
    if (channel == null) {
      channel = MethodChannel('plugin.gem.com/cocos_view_$cocosId');

      channel.setMethodCallHandler(
          (MethodCall call) => _handleMethodCall(call, cocosId));
      _channels[cocosId] = channel;
    }
    return channel;
  }

  /// Initializes the platform interface with [id].
  ///
  /// This method is called when the plugin is first initialized.
  @override
  Future<void> init(int cocosId) {
    MethodChannel channel = ensureChannelInitialized(cocosId);
    return channel.invokeMethod<void>('cocos#waitForCocos');
  }

  /// Dispose of the native resources.
  @override
  Future<void> dispose({int? cocosId}) async {
    try {
      if (cocosId != null) await channel(cocosId).invokeMethod('cocos#dispose');
    } catch (e) {
      // ignore
    }
  }

  // The controller we need to broadcast the different events coming
  // from handleMethodCall.
  //
  // It is a `broadcast` because multiple controllers will connect to
  // different stream views of this Controller.
  final StreamController<CocosEvent> _cocosStreamController =
      StreamController<CocosEvent>.broadcast();

  // Returns a filtered view of the events in the _controller, by cocosId.
  Stream<CocosEvent> _events(int cocosId) =>
      _cocosStreamController.stream.where((event) => event.cocosId == cocosId);

  Future<dynamic> _handleMethodCall(MethodCall call, int cocosId) async {
    switch (call.method) {
      case "events#onCocosMessage":
        _cocosStreamController.add(CocosMessageEvent(cocosId, call.arguments));
        break;
      case "events#onCocosUnloaded":
        _cocosStreamController.add(CocosLoadedEvent(cocosId, call.arguments));
        break;
      case "events#onCocosSceneLoaded":
        _cocosStreamController.add(CocosSceneLoadedEvent(
            cocosId, SceneLoaded.fromMap(call.arguments)));
        break;
      case "events#onCocosCreated":
        _cocosStreamController.add(CocosCreatedEvent(cocosId, call.arguments));
        break;
      default:
        throw UnimplementedError("Unimplemented ${call.method} method");
    }
  }

  @override
  Future<bool?> isPaused({required int cocosId}) async {
    return await channel(cocosId).invokeMethod('cocos#isPaused');
  }

  @override
  Future<bool?> isReady({required int cocosId}) async {
    return await channel(cocosId).invokeMethod('cocos#isReady');
  }

  @override
  Future<bool?> isLoaded({required int cocosId}) async {
    return await channel(cocosId).invokeMethod('cocos#isLoaded');
  }

  @override
  Future<bool?> inBackground({required int cocosId}) async {
    return await channel(cocosId).invokeMethod('cocos#inBackground');
  }

  @override
  Future<bool?> createCocosPlayer({required int cocosId}) async {
    return await channel(cocosId).invokeMethod('cocos#createPlayer');
  }

  @override
  Stream<CocosMessageEvent> onCocosMessage({required int cocosId}) {
    return _events(cocosId).whereType<CocosMessageEvent>();
  }

  @override
  Stream<CocosLoadedEvent> onCocosUnloaded({required int cocosId}) {
    return _events(cocosId).whereType<CocosLoadedEvent>();
  }

  @override
  Stream<CocosCreatedEvent> onCocosCreated({required int cocosId}) {
    return _events(cocosId).whereType<CocosCreatedEvent>();
  }

  @override
  Stream<CocosSceneLoadedEvent> onCocosSceneLoaded({required int cocosId}) {
    return _events(cocosId).whereType<CocosSceneLoadedEvent>();
  }

  @override
  Widget buildViewWithTextDirection(
    int creationId,
    PlatformViewCreatedCallback onPlatformViewCreated, {
    required TextDirection textDirection,
    Set<Factory<OneSequenceGestureRecognizer>>? gestureRecognizers,
    Map<String, dynamic> cocosOptions = const <String, dynamic>{},
    bool? useAndroidViewSurf,
    bool? height,
    bool? width,
    bool? cocosWebSource,
    String? cocosSrcUrl,
  }) {
    final String _viewType = 'plugin.gem.com/cocos_view';

    if (useAndroidViewSurf != null) useAndroidViewSurface = useAndroidViewSurf;

    final Map<String, dynamic> creationParams = cocosOptions;

    if (defaultTargetPlatform == TargetPlatform.windows) {
      return WindowsCocosWidgetView();
    }

    if (defaultTargetPlatform == TargetPlatform.android) {
      if (!useAndroidViewSurface) {
        return AndroidView(
          viewType: _viewType,
          onPlatformViewCreated: onPlatformViewCreated,
          gestureRecognizers: gestureRecognizers,
          creationParams: creationParams,
          creationParamsCodec: const StandardMessageCodec(),
          hitTestBehavior: PlatformViewHitTestBehavior.opaque,
          layoutDirection: TextDirection.ltr,
        );
      }

      return PlatformViewLink(
        viewType: _viewType,
        surfaceFactory: (
          BuildContext context,
          PlatformViewController controller,
        ) {
          return AndroidViewSurface(
            controller: controller as AndroidViewController,
            gestureRecognizers: gestureRecognizers ??
                const <Factory<OneSequenceGestureRecognizer>>{},
            hitTestBehavior: PlatformViewHitTestBehavior.opaque,
          );
        },
        onCreatePlatformView: (PlatformViewCreationParams params) {
          final controller = PlatformViewsService.initExpensiveAndroidView(
            id: params.id,
            viewType: _viewType,
            layoutDirection: TextDirection.ltr,
            creationParams: creationParams,
            creationParamsCodec: const StandardMessageCodec(),
            onFocus: () => params.onFocusChanged(true),
          );

          controller
            ..addOnPlatformViewCreatedListener(params.onPlatformViewCreated)
            ..addOnPlatformViewCreatedListener(onPlatformViewCreated)
            ..create();
          return controller;
        },
      );
    } else if (defaultTargetPlatform == TargetPlatform.iOS) {
      return UiKitView(
        viewType: _viewType,
        onPlatformViewCreated: onPlatformViewCreated,
        gestureRecognizers: gestureRecognizers,
        creationParams: creationParams,
        creationParamsCodec: const StandardMessageCodec(),
      );
    }
    return Text(
        '$defaultTargetPlatform is not yet supported by the cocos player plugin');
  }

  @override
  Widget buildView(
    int creationId,
    PlatformViewCreatedCallback onPlatformViewCreated, {
    Map<String, dynamic> cocosOptions = const {},
    Set<Factory<OneSequenceGestureRecognizer>>? gestureRecognizers,
    bool? useAndroidViewSurf,
    String? cocosSrcUrl,
  }) {
    return buildViewWithTextDirection(
      creationId,
      onPlatformViewCreated,
      textDirection: TextDirection.ltr,
      gestureRecognizers: gestureRecognizers,
      cocosOptions: cocosOptions,
      useAndroidViewSurf: useAndroidViewSurf,
      cocosSrcUrl: cocosSrcUrl,
    );
  }

  @override
  Future<void> postMessage({
    required int cocosId,
    required String gameObject,
    required String methodName,
    required String message,
  }) async {
    await channel(cocosId).invokeMethod('cocos#postMessage', <String, dynamic>{
      'gameObject': gameObject,
      'methodName': methodName,
      'message': message,
    });
  }

  @override
  Future<void> postJsonMessage({
    required int cocosId,
    required String gameObject,
    required String methodName,
    required Map message,
  }) async {
    await channel(cocosId).invokeMethod('cocos#postMessage', <String, dynamic>{
      'gameObject': gameObject,
      'methodName': methodName,
      'message': json.encode(message),
    });
  }

  @override
  Future<void> pausePlayer({required int cocosId}) async {
    await channel(cocosId).invokeMethod('cocos#pausePlayer');
  }

  @override
  Future<void> resumePlayer({required int cocosId}) async {
    await channel(cocosId).invokeMethod('cocos#resumePlayer');
  }

  @override
  Future<void> openInNativeProcess({required int cocosId}) async {
    await channel(cocosId).invokeMethod('cocos#openInNativeProcess');
  }

  @override
  Future<void> unloadPlayer({required int cocosId}) async {
    await channel(cocosId).invokeMethod('cocos#unloadPlayer');
  }

  @override
  Future<void> quitPlayer({required int cocosId}) async {
    await channel(cocosId).invokeMethod('cocos#quitPlayer');
  }
}
